/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QtWidgets/QPushButton>
#include <QtWidgets/QDialog>
#include <QtWidgets/QGroupBox>
#include "controller.h"

// TODO use sizers

DiotSlot::DiotSlot(QDialog* parent, DiotController& controller, int num, int idx)
    : controller(controller), inputs(num, true), outputs(num, false), slot_nr(idx), io_count(num)
{
    groupBox = new QGroupBox(parent);
    groupBox->setGeometry(QRect(20 + idx * 169, 20, 150, 40 + io_count * 30));
    groupBox->setTitle(QString("Slot %1").arg(idx + 1));

    for(int i = 0; i < io_count; ++i) {
        QPushButton* state = new QPushButton(groupBox);
        state->setGeometry(QRect(20, 30 + i * 30, 30, 30));
        state->setText(QString("%1").arg((i + 1) + idx * num));
        btn_state.push_back(state);

        QPushButton* test = new QPushButton(groupBox);
        test->setGeometry(QRect(50, 30 + i * 30, 80, 30));
        test->setText("Test");
        QObject::connect(test, &QPushButton::clicked, [=](){ action_click(idx, i); });
        btn_test.push_back(test);
    }

    QObject::connect(this, &DiotSlot::InputChanged, [=](){ update_inputs(); });
    QObject::connect(this, &DiotSlot::OutputChanged,
            &controller, &DiotController::SetOutput);
    update_inputs();
}


void DiotSlot::SetInput(unsigned int state) {
    if(int_to_vec(state, inputs))
        emit InputChanged(slot_nr, state);
}


void DiotSlot::SetOutput(unsigned int state) {
    if(int_to_vec(state, outputs))
        emit OutputChanged(slot_nr, state);
}


void DiotSlot::action_click(int slot, int io) {
    qDebug() << "Button test clicked - slot(" << slot << ") io(" << io << ")";
    outputs[io] = !outputs[io];
    emit OutputChanged(slot_nr, GetOutput());
}


void DiotSlot::update_inputs() {
    for(int i = 0; i < io_count; ++i) {
        if(inputs[i]) {
            btn_state[i]->setStyleSheet("color: green");
        } else {
            btn_state[i]->setStyleSheet("color: red");
        }
    }
}


unsigned int DiotSlot::vec_to_int(const std::vector<bool>& vec) {
    unsigned int res = 0;
    int mask = 1;

    for(bool state : vec) {
        if(state)
            res |= mask;

        mask <<= 1;
    }

    return res;
}


bool DiotSlot::int_to_vec(unsigned int integer, std::vector<bool>& vec)
{
    bool changed = false;

    for(unsigned int i = 0; i < vec.size(); ++i) {
        bool new_state = (integer & 1);

        if(vec[i] != new_state) {
            vec[i] = new_state;
            changed = true;
        }

        integer >>= 1;
    }

    return changed;
}


DiotController::DiotController(QDialog* dialog, int address, int slots_nr) {
    // Create widgets
    for(int i = 0; i < slots_nr; ++i) {
        io_slots.emplace_back(new DiotSlot(dialog, *this, 16, i));
    }

    // Configure MasterFIP
    mfip = MasterFIP::OpenByLUN(0);

    // Hardware and software config for MasterFIP
    memset(&hw_cfg, 0, sizeof(mstrfip_hw_cfg));
    hw_cfg.enable_ext_trig = 0;
    hw_cfg.enable_int_trig = 0;
    hw_cfg.enable_ext_trig_term = 0;
    hw_cfg.turn_around_ustime = 14;         // 14 us corresponds to 1 Mbit/s

    memset(&sw_cfg, 0, sizeof(mstrfip_sw_cfg));
    hw_cfg.enable_ext_trig = 0;
    sw_cfg.irq_thread_prio = 50;
    sw_cfg.diag_thread_prio = 20;

    // Cycle configuration (60 ms duration)
    cycle.reset(new CycleSimple(*mfip, 60000));
    crate.reset(new DIOT_Crate(address, slots_nr));
    crate->AddVariables(cycle->GetPeriodicVarWindow());

    crate->SetConsVarCb([&](nanoFIP &nf) {
        // Report the inputs state
        DIOT_Crate* crate = static_cast<DIOT_Crate*>(&nf);

        for(int i = 0; i < crate->GetCardCount(); ++i) {
            emit InputChanged(i, crate->GetSlotInputs(i));
        }
    } );

    QObject::connect(this, &DiotController::InputChanged,
            this, &DiotController::UpdateInput);
}


void DiotController::Start() {
    mfip->Configure(hw_cfg, sw_cfg);
    mfip->StartCycle(*cycle);
}


void DiotController::UpdateInput(int slot, unsigned int state)
{
    qDebug() << "received input update: slot " << slot << "=" << state;
    io_slots[slot]->SetInput(state);
}


void DiotController::SetOutput(int slot, unsigned int state)
{
    qDebug() << "sending output update: slot " << slot << "=" << state;
    crate->SetSlotOutputs(slot, state);
}
