/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DIOT_CRATE_H
#define DIOT_CRATE_H

#include <nanofip.h>

/**
 * DIOT_Crate represents a CompactPCI-Serial crate stuffed with digital
 * input/output cards (EDA-03828) managed via nanoFIP.
 */
class DIOT_Crate : public nanoFIP
{
public:
    /**
     * Default constructor.
     * @param addr is the FIP agent address.
     * @param cards is the number of I/O cards in the crate.
     */
    DIOT_Crate(int addr, int cards)
        : nanoFIP(addr, VAR_SIZE), card_count(cards) {}

    /**
     * Change an output state.
     * @param slot is the card number (indexing from 0).
     * @param idx is the output number (0 to IO_PER_SLOT - 1 range).
     * @param state is the new output state.
     */
    void SetOutput(int slot, int idx, bool state);

    /**
     * Return an input state.
     * @param slot is the card number (indexing from 0).
     * @param idx is the input number (0 to IO_PER_SLOT - 1 range).
     * @return True when the input is high, false otherwise.
     */
    bool GetInput(int slot, int idx);

    /**
     * Change state of all outputs of a card in a particular slot.
     * @param slot is the card number (indexing from 0).
     * @param state is the new outputs state. The LSB corresponds to I/O number 0.
     */
    void SetSlotOutputs(int slot, uint32_t state);

    /**
     * Return state of all inputs of a card in a particular slot.
     * @param slot is the card number (indexing from 0).
     * @return the inputs state. The LSB corresponds to I/O number 0.
     */
    uint32_t GetSlotInputs(int slot);

    /**
     * Return the number of I/O cards installed in the crate.
     * Note there is no query mechanism involved, it is just the number
     * that has been passed to the constructor.
     */
    int GetCardCount() const
    {
        return card_count;
    }

private:
    ///> Return variable offset corresponding to a particular I/O and slot.
    static int getVarOffset(int slot, int idx)
    {
        return slot * (IO_PER_SLOT / 8) + idx / 8;
    }

    ///> Calculates bitmask for particular I/O.
    static uint8_t getIOMask(int idx)
    {
        return (1 << (idx % 8));
    }

    ///> Produced/consumed variable size (expressed in bytes)
    static constexpr int VAR_SIZE = 124;

    ///> Number of I/Os per slot (in the memory map)
    static constexpr int IO_PER_SLOT = 32;

    ///> Number of I/O cards installed in the crate
    const int card_count;
};

#endif /* DIOT_CRATE_H */
